//
//    StickyKeys for EPOC
//
//    Copyright (C) 2000  Marc Boogaards
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include "StickyKeys.h"

#pragma data_seg(".E32_UID")
__WINS_UID(KExecutableImageUidValue,0x00000000,0x00000000)
#pragma data_seg()

///////////////////////////////////////////////////////////////////////////////
//	CKeyState
//
//	Implements a simple state machine to keep track of 'sticky state' for a
//	certain modifier key, in response to key presses. The state diagram is
//	included in KeyState.bmp. Implementation of the state changes is rather
//	blunt but effective.
///////////////////////////////////////////////////////////////////////////////

// Process a key event. In stead of passing the actualy event to this function,
// one of KeyDown, KeyUp, OtherDown and OtherUp should be passed. KeyDown/KeyUp
// for the modifier key, OtherDown/OtherUp for any other key.
void CKeyState::ProcessKey(int Event)
{
	int StateBefore = iState;

	switch (iState)
	{
	case S0:
		if (Event == KeyDown)
			iState = S1;
		break;

	case S1:
		if (Event == KeyUp)
			iState = S2;
		if (Event == OtherDown)
			iState = S5;
		break;

	case S2:
		if (Event == KeyDown)
			iState = S3;
		if (Event == OtherDown)
			iState = S4;
		break;

	case S3:
		if (Event == KeyUp)
			iState = S0;
		break;

	case S4:
		if (Event == OtherUp)
			iState = S0;
		break;

	case S5:
		if (Event == KeyUp)
			iState = S0;
		break;

	default:
		break;
	}
	// set the changed flag
	iChanged = (iState & MASK_STICKY) != (StateBefore & MASK_STICKY);
}

// returns whether the key is currently down (regardless
// of the current sticky state)
int CKeyState::Down() const
{
	return (iState & MASK_KEYDOWN) != 0;
}

// returns whether the key is currently sticky
int CKeyState::Sticky() const
{
	return (iState & MASK_STICKY) != 0;
}

// returns whether the sticky state was changed since the last
// time this query was done
int CKeyState::Changed()
{
	int rv = iChanged;
	iChanged = FALSE;
	return rv;
}

void CKeyState::Reset()
{
	iState = S0;
	iChanged = FALSE;
}

CKeyState::CKeyState()
{
	Reset();
}

CKeyState::~CKeyState()
{
}

///////////////////////////////////////////////////////////////////////////////
//	CWindow
//
//	Implements the small message window in the lower left corner of the
//	screen showing the current key state. This window works solely on the
//	Windows Server, and is independent of EIKON.
///////////////////////////////////////////////////////////////////////////////

// the border width is the amount of black around the white border
// the white border is one pixel wide
#define BORDER_WIDTH 2

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CWindow::CWindow(CWsClient* aClient) : iClient(aClient)
{

}

CWindow::~CWindow()
{
	iClient->iScreen->ReleaseFont(iFont);
	iWindow.Close(); // close our window
}

void CWindow::ConstructL (const TMsg &aText, TInt nFontSize)
{
	_LIT(KFont,"Arial");

	FontSize = nFontSize;
	TFontSpec spec(KFont, FontSize);

	iWindow = RWindow(iClient->iWs); // use app's session to window server
	User::LeaveIfError(iWindow.Construct(iClient->iGroup, (TUint32)this));

	// make sure the message window is always on top
	// (this is one of the place where this happens, see also
	// CWsClient::ConstructL, and CWsClient::RunL)
	iWindow.SetOrdinalPosition(ECoeWinPriorityAlwaysAtFront);

	iClient->iScreen->GetNearestFontInTwips(iFont, spec); // get the system message font

	SetText(aText);

	iWindow.Activate(); // window is now active
}

void CWindow::SetText(const TMsg &aText)
{
	TInt textHeight, textWidth;
	TRect iRect; // rectangle re owning window

	aString.Copy(aText);
	textHeight = iFont->HeightInPixels();

	textWidth = iFont->TextWidthInPixels(aText);
	iRect = TRect(TPoint(0,0), TPoint(textWidth + 4 + (4*BORDER_WIDTH), textHeight + 4 + (4*BORDER_WIDTH)));
	iRect.Move(10, iClient->iScreen->SizeInPixels().iHeight - iRect.Height() - 10);

	iBaselineOffset = (textHeight + iRect.Height())/2 - 1 ; // 1/2 font ht below halfway down box

	iWindow.SetExtent(iRect.iTl, iRect.Size()); // set extent relative to group coords
	iWindow.SetOrdinalPosition(ECoeWinPriorityAlwaysAtFront); // always on top
	iWindow.Invalidate();
}

RWindow& CWindow::Window()
{
	return iWindow;
}

CWindowGc* CWindow::SystemGc()
{
	return iClient->iGc;
}

/****************************************************************************\
|	Function:	CWindow::Draw
|	Purpose:	Redraws the contents of CWindow within a given
|				rectangle.
|	Input:		aRect	Rectangle that needs redrawing
|	Output:		None
\****************************************************************************/
void CWindow::Draw(const TRect& aRect)
{
	// Drawing to a window is done using functions supplied by
	// the graphics context (CWindowGC), not the window.
	CWindowGc* gc=SystemGc(); // get a gc
	gc->SetClippingRect(aRect); // clip outside this rect
	gc->Clear(aRect); // clear
	TSize size=iWindow.Size();
	TInt width=size.iWidth;
	TInt height=size.iHeight;

    // draw some text centered in a box, (margin is zero)
	gc->SetBrushStyle(CGraphicsContext::ESolidBrush);
	gc->SetBrushColor(KRgbBlack);
	gc->SetPenColor(KRgbWhite);
	gc->UseFont(iFont);
	gc->DrawText(aString, aRect, iBaselineOffset, CGraphicsContext::ECenter, 0);
	gc->DiscardFont();

	// Draw a square border
	gc->DrawLine(TPoint(BORDER_WIDTH, BORDER_WIDTH), TPoint(BORDER_WIDTH, height - (2*BORDER_WIDTH)));
	gc->DrawLine(TPoint(BORDER_WIDTH, height - (2*BORDER_WIDTH)), TPoint(width - (2*BORDER_WIDTH), height - (2*BORDER_WIDTH)));
	gc->DrawLine(TPoint(width - (2*BORDER_WIDTH), height - (2*BORDER_WIDTH)), TPoint(width - (2*BORDER_WIDTH), BORDER_WIDTH));
	gc->DrawLine(TPoint(width - (2*BORDER_WIDTH), BORDER_WIDTH), TPoint(BORDER_WIDTH, BORDER_WIDTH));

}

/****************************************************************************\
|	Function:	CWindow::HandlePointerEvent
|	Purpose:	Handles pointer events for CWindow.  Doesn't do
|				anything except get the co-ordinates where the pointer
|				event occurred.
|	Input:		aPointerEvent	The pointer event
|	Output:		None
\****************************************************************************/

void CWindow::HandlePointerEvent(TPointerEvent&)
{
}

///////////////////////////////////////////////////////////////////////////////
//	CWsRedrawer
//
//	Helper for CWindow
///////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CWsRedrawer::CWsRedrawer() : CActive(CActive::EPriorityLow)
{

}

CWsRedrawer::~CWsRedrawer()
{
	Cancel();
}

void CWsRedrawer::ConstructL(CWsClient* aClient)
{
	iClient=aClient; // remember WsClient that owns us
	CActiveScheduler::Add(this); // add ourselves to the scheduler
	IssueRequest(); // issue request to draw
}

void CWsRedrawer::IssueRequest()
{
	iClient->iWs.RedrawReady(&iStatus);
	SetActive();
}

void CWsRedrawer::DoCancel()
{
	iClient->iWs.RedrawReadyCancel();
}

void CWsRedrawer::RunL()
{
	// find out what needs to be done
	TWsRedrawEvent redrawEvent;
    iClient->iWs.GetRedraw(redrawEvent); // get event
	CWindow* window=(CWindow*)(redrawEvent.Handle()); // get window
	if (window)
	{
		TRect rect=redrawEvent.Rect(); // and rectangle that needs redrawing
		// now do drawing
		iClient->iGc->Activate(window->Window());
		window->Window().BeginRedraw(rect);
		window->Draw(rect);
		window->Window().EndRedraw();
		iClient->iGc->Deactivate();
	}
	// maintain outstanding request
	IssueRequest();
}

///////////////////////////////////////////////////////////////////////////////
//	CWsClient
//
//	Implements the Windows Server client, which creates a scheduler and
//	installs a basically the hook that monitors all key events
///////////////////////////////////////////////////////////////////////////////

// Handy macro that seets the modifier state for a specified key
// Modifier state is set for both Left and Right key, AND for the
// 'generic' modifier. This specific combination works, others seem
// to fail.
#define SETMODIFIER(KeyName,On)	{\
		iWs.SetModifierState(EModifierLeft ##KeyName, On?ETurnOnModifier:ETurnOffModifier);\
		iWs.SetModifierState(EModifierRight ##KeyName, On?ETurnOnModifier:ETurnOffModifier);\
		iWs.SetModifierState(EModifier ##KeyName, On?ETurnOnModifier:ETurnOffModifier);\
	}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CWsClient::CWsClient() : CActive(CActive::EPriorityStandard)
{

}

CWsClient::~CWsClient()
{
	// neutralize us as an active object
	Deque(); // cancels and removes from scheduler
	// get rid of scheduler and all attached objects
	delete CActiveScheduler::Current(); // delete the scheduler
	delete iScreen;
	delete iRedrawer;
	for (int i = 0; i <= 255; i++)
		iGroup.CancelCaptureKeyUpAndDowns(hCaptureCtrl[i]);
	// destroy window group
	iGroup.Close();
	// finish with window server
	iWs.Close();
	iGlobalMutex.Close();
}

void CWsClient::ConstructL()
{
	TBuf<100> ThreadName=_S(APP_NAME);
	RThread iThread;
	TThreadId aThreadID;

	// if the global mutex alredy exists, quit
	if (iGlobalMutex.OpenGlobal(ThreadName) == KErrNone)
	{
		iGlobalMutex.Close();
		User::Exit(0);
		return;
	}
	// create the global mutex
	iGlobalMutex.CreateGlobal(ThreadName);

	// construct an active scheduler and add ourselves to it
	CActiveScheduler::Install(new (ELeave) CActiveScheduler);
	CActiveScheduler::Add(this);
	// connect a session to the window server
	User::LeaveIfError(iWs.Connect());

	// construct our one and only window group
	iGroup=RWindowGroup(iWs);
	User::LeaveIfError(iGroup.Construct(2));	// "2" is the handle we're giving
												// to the window group - meaningless
												// in this case as we only have one
												// window group.

	// rename the main (only) thread, so it shows nicely in the
	// Open Files dialog
	iWs.GetWindowGroupClientThreadId(iGroup.Identifier(), aThreadID);
	if (iThread.Open(aThreadID) == KErrNone)
	{
		iThread.Rename(ThreadName);
	}
	else
	{
		User::Abort();
	}

	// set the window group to be topmost, to make sure the message
	// window is always on top (this is one of the place where something
	// like this happens, see also CWindow::ConstructL, and CWsClient::RunL)
	iWs.SetWindowGroupOrdinalPosition(iGroup.Identifier(), 0);
	iGroup.EnableReceiptOfFocus(EFalse);	// do not accept keyboard focus
	iGroup.SetOrdinalPriorityAdjust(ECoeWinPriorityAlwaysAtFront);

	// construct screen device and graphics context
	iScreen=new (ELeave) CWsScreenDevice(iWs); // make device for this session
	User::LeaveIfError(iScreen->Construct()); // and complete its construction
	User::LeaveIfError(iScreen->CreateContext(iGc)); // create graphics context
	// construct redrawer
	iRedrawer=new (ELeave) CWsRedrawer;
	iRedrawer->ConstructL(this);

	// these are the events we're after, and we want them always!

	for (int i = 0; i <= 255; i++)
		hCaptureCtrl[i] = iGroup.CaptureKeyUpAndDowns(i, 0, 0);

	iGroup.EnableErrorMessages(EEventControlAlways);
	iGroup.EnableOnEvents(EEventControlAlways);

	iGroup.EnableModifierChangedEvents(EAllModifiers, EEventControlAlways);
	// request first event and start scheduler
	IssueRequest();
	CActiveScheduler::Start(); // start the active scheduler
}

void CWsClient::Exit()
{
	// reset modifier states
	if (CtrlState.Sticky())
		SETMODIFIER(Ctrl,FALSE);

	if (ShiftState.Sticky())
		SETMODIFIER(Shift,FALSE);

	if (FuncState.Sticky())
		SETMODIFIER(Func,FALSE);

	CActiveScheduler::Stop();
}

void CWsClient::IssueRequest()
{
	iWs.EventReady(&iStatus); // request an event

	SetActive(); // so we're now active
}

void CWsClient::DoCancel()
{
	iWs.EventReadyCancel(); // cancel event request
}

void CWsClient::RunL()
{
	// get the event
	iWs.GetEvent(iWsEvent);
	TInt eventType=iWsEvent.Type();
	// take action on it
	switch (eventType)
	{
		// window-group related event types
	case EEventNull:
	case EEventKey:
		break;

	case EEventKeyDown:
		{
			// process the key down event. that is: send it to the
			// relevant state machines
			TKeyEvent& key=*iWsEvent.Key(); // get key event

			switch (key.iScanCode)
			{
			case EStdKeyLeftCtrl:
			case EStdKeyRightCtrl:
				CtrlState.ProcessKey(CKeyState::KeyDown);
				break;

			case EStdKeyLeftShift:
			case EStdKeyRightShift:
				ShiftState.ProcessKey(CKeyState::KeyDown);
				break;

			case EStdKeyLeftFunc:
			case EStdKeyRightFunc:
				FuncState.ProcessKey(CKeyState::KeyDown);
				break;

			case 'Q':
			case 'q':
				if (FuncState.Down())
				{
					// Fn-Q is pressed: quit
					// reset the modifier states. if we don't EPOC
					// will start acting up, and the modifiers (specifically
					// Ctrl may be inverted)
					if (CtrlState.Sticky())
						SETMODIFIER(Ctrl,FALSE);

					if (FuncState.Sticky())
						SETMODIFIER(Shift,FALSE);

					if (FuncState.Sticky())
						SETMODIFIER(Func,FALSE);

					Exit();
				}
				// fall through!
			default:
				CtrlState.ProcessKey(CKeyState::OtherDown);
				ShiftState.ProcessKey(CKeyState::OtherDown);
				FuncState.ProcessKey(CKeyState::OtherDown);
				break;
			}
			break;
		}

	case EEventKeyUp:
		{
			// process the key up event. again this means
			// having it handled by the relevant state machine
			TKeyEvent& key = *iWsEvent.Key(); // get key event

			switch (key.iScanCode)
			{
			case EStdKeyLeftCtrl:
			case EStdKeyRightCtrl:
				CtrlState.ProcessKey(CKeyState::KeyUp);
				break;

			case EStdKeyLeftShift:
			case EStdKeyRightShift:
				ShiftState.ProcessKey(CKeyState::KeyUp);
				break;

			case EStdKeyLeftFunc:
			case EStdKeyRightFunc:
				FuncState.ProcessKey(CKeyState::KeyUp);
				break;

			default:
				CtrlState.ProcessKey(CKeyState::OtherUp);
				ShiftState.ProcessKey(CKeyState::OtherUp);
				FuncState.ProcessKey(CKeyState::OtherUp);
				break;
			}
			break;
		}

	case EEventModifiersChanged:
	case EEventFocusLost:
		// make sure our window group comes out on top, otherwise
		// the nice message window will not be visible...
		// (see also CWsClient::ConstructL, and CWindow::ContructL)
		iWs.SetWindowGroupOrdinalPosition(iGroup.Identifier(), 0);
		break;

	case EEventFocusGained:
	case EEventSwitchOn:
	case EEventPassword:
	case EEventWindowGroupsChanged:
	case EEventErrorMessage:
	case EEventPointer:
	case EEventPointerEnter:
	case EEventPointerExit:
	case EEventPointerBufferReady:
	case EEventDragDrop:
		break;

	default:
		Exit();
		break;
	}

	// set the modifier state according to what the state machines
	// think it should be
	if (CtrlState.Changed())
		SETMODIFIER(Ctrl, CtrlState.Sticky());

	if (ShiftState.Changed())
		SETMODIFIER(Shift, ShiftState.Sticky());

	if (FuncState.Changed())
		SETMODIFIER(Func, FuncState.Sticky());

	// and on the key up event, display or adjust the message
	if (eventType == EEventKeyUp)
	{
		TMsg mCtrl = _S("Ctrl");
		TMsg mFn = _S("Fn");
		TMsg mShift = _S("Shift");
		TMsg mPlus = _S("+");
		TInt fLen = 0;
		TMsg msg;

		msg.Zero();

		if (CtrlState.Sticky())
		{
			msg.Append(mCtrl);
			fLen = 1;
		}

		if (ShiftState.Sticky())
		{
			if (fLen)
			{
				msg.Append(mPlus);
			}
			msg.Append(mShift);
			fLen = 1;
		}

		if (FuncState.Sticky())
		{
			if (fLen)
			{
				msg.Append(mPlus);
			}
			msg.Append(mFn);
			fLen = 1;
		}

		if (msg.Length())
		{
			if (iMsgBox)
			{
				iMsgBox->SetText(msg);
			}
			else
			{
				iMsgBox = new CWindow(this);
				iMsgBox->ConstructL(msg, 150);
			}
		}
		else
		{
			if (iMsgBox)
			{
				delete iMsgBox;
				iMsgBox = NULL;
			}
		}
	}
	IssueRequest(); // maintain outstanding request
}


///////////////////////////////////////////////////////////////////////////////
//     globals
///////////////////////////////////////////////////////////////////////////////

void RunStickyKeys()
{
	User::SetJustInTime(EFalse);
	// make new client
	CWsClient* client=new (ELeave) CWsClient; // allocate new client
	CleanupStack::PushL(client); // push, just in case
	client->ConstructL(); // construct and run
	CleanupStack::PopAndDestroy(); // destruct
}


#if defined(__WINS__)
EXPORT_C TInt WinsMain(TAny *)
#endif
#if defined(__EPOC32__)
GLDEF_C TInt E32Main()
#endif
{
	_LIT(KPanicMsg,"EPOC32EX");
	CTrapCleanup* cleanup=CTrapCleanup::New(); // get clean-up stack
	TRAPD(error, RunStickyKeys()); // more initialization
	__ASSERT_ALWAYS(!error,User::Panic(KPanicMsg,error));
	delete cleanup; // destroy clean-up stack
	return KErrNone;
}

GLDEF_C TInt E32Dll(TDllReason)
{
	return KErrNone;
}
